# 自己平常写代码时为了方便实现某些功能而写的工具函数
import pandas as pd
import torch
import os
import numpy as np
import random
import time
from tqdm import tqdm
from Bio import SeqIO


# 统计时间的函数
class TimeCounter:
    def __init__(self, text, verbose=True):
        self.text = text
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()
        if self.verbose:
            print(self.text, flush=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        end = time.time()
        t = end - self.start
        if self.verbose:
            print(f"Finished. The time is {t:.2f}s.\n", flush=True)


# 进度条
def progress_bar(now: int, total: int, desc: str = '', end='\n'):
    length = 50
    now = now if now <= total else total
    num = now * length // total
    progress_bar = '[' + '#' * num + '_' * (length - num) + ']'
    display = f'{desc:} {progress_bar} {int(now/total*100):02d}% {now}/{total}'

    print(f'\r\033[31m{display}\033[0m', end=end, flush=True)


# 设定种子
def setup_seed(seed):
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.deterministic = True


# 随机种子
def random_seed():
    torch.seed()
    torch.cuda.seed()
    np.random.seed()
    random.seed()
    torch.backends.cudnn.deterministic = False


if __name__ == '__main__':
    # file_names = [f"{i}_{i+100}" for i in range(0, 1100, 100)]
    # files = []
    # path = "/sujin/dataset/uniclust30"
    # for name in file_names:
    #     files.append(open(f"{path}/UniRef30_2020_06_{name}", 'w'))
    #
    # fasta_path = "/sujin/dataset/uniclust30/UniRef30_2020_06.faa"
    # with open(fasta_path, 'r') as r:
    #     desc = r.readline()
    #     while desc:
    #         seq = r.readline()
    #         if len(seq) - 1 > 1022:
    #             desc = r.readline()
    #             continue
    #
    #         index = int((len(seq) - 1) / 100)
    #         files[index].write(desc)
    #         files[index].write(seq)
    #
    #         desc = r.readline()
    
    a = 1
    print(a)
    


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math
import torch
import torch.nn as nn


class RowSelfAttention(nn.Module):
    """Compute self-attention over rows of a 2D input."""

    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        max_tokens_per_msa: int = 2 ** 16,
    ):
        super().__init__()
        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = embed_dim // num_heads
        self.scaling = self.head_dim ** -0.5
        self.max_tokens_per_msa = max_tokens_per_msa
        self.attn_shape = "hnij"

        self.k_proj = nn.Linear(embed_dim, embed_dim)
        self.v_proj = nn.Linear(embed_dim, embed_dim)
        self.q_proj = nn.Linear(embed_dim, embed_dim)

        self.out_proj = nn.Linear(embed_dim, embed_dim)
        self.dropout_module = nn.Dropout(dropout)

    def align_scaling(self, q):
        num_rows = q.size(0)
        return self.scaling / math.sqrt(num_rows)

    def _batched_forward(
        self,
        x,
        self_attn_mask=None,
        self_attn_padding_mask=None,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        max_rows = max(1, self.max_tokens_per_msa // num_cols)
        attns = 0
        scaling = self.align_scaling(x)
        for start in range(0, num_rows, max_rows):
            attn_weights = self.compute_attention_weights(
                x[start : start + max_rows],
                scaling,
                self_attn_mask=self_attn_mask,
                self_attn_padding_mask=self_attn_padding_mask[:, start : start + max_rows]
                if self_attn_padding_mask is not None
                else None,
            )
            attns += attn_weights
        attn_probs = attns.softmax(-1)
        attn_probs = self.dropout_module(attn_probs)

        outputs = []
        for start in range(0, num_rows, max_rows):
            output = self.compute_attention_update(x[start : start + max_rows], attn_probs)
            outputs.append(output)

        output = torch.cat(outputs, 0)
        return output, attn_probs

    def compute_attention_weights(
        self,
        x,
        scaling: float,
        self_attn_mask=None,
        self_attn_padding_mask=None,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        q = self.q_proj(x).view(num_rows, num_cols, batch_size, self.num_heads, self.head_dim)
        k = self.k_proj(x).view(num_rows, num_cols, batch_size, self.num_heads, self.head_dim)
        q *= scaling
        if self_attn_padding_mask is not None:
            # Zero out any padded aligned positions - this is important since
            # we take a sum across the alignment axis.
            q *= 1 - self_attn_padding_mask.permute(1, 2, 0).unsqueeze(3).unsqueeze(4).to(q)

        attn_weights = torch.einsum(f"rinhd,rjnhd->{self.attn_shape}", q, k)

        if self_attn_mask is not None:
            raise NotImplementedError
            # Mask Size: [B x R x C], Weights Size: [H x B x C x C]

        if self_attn_padding_mask is not None:
            attn_weights = attn_weights.masked_fill(
                self_attn_padding_mask[:, 0].unsqueeze(0).unsqueeze(2),
                -10000,
            )

        return attn_weights

    def compute_attention_update(
        self,
        x,
        attn_probs,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        v = self.v_proj(x).view(num_rows, num_cols, batch_size, self.num_heads, self.head_dim)
        context = torch.einsum(f"{self.attn_shape},rjnhd->rinhd", attn_probs, v)
        context = context.contiguous().view(num_rows, num_cols, batch_size, embed_dim)
        output = self.out_proj(context)
        return output

    def forward(
        self,
        x,
        self_attn_mask=None,
        self_attn_padding_mask=None,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        if (num_rows * num_cols > self.max_tokens_per_msa) and not torch.is_grad_enabled():
            return self._batched_forward(x, self_attn_mask, self_attn_padding_mask)
        else:
            scaling = self.align_scaling(x)
            attn_weights = self.compute_attention_weights(
                x, scaling, self_attn_mask, self_attn_padding_mask
            )
            attn_probs = attn_weights.softmax(-1)
            attn_probs = self.dropout_module(attn_probs)
            output = self.compute_attention_update(x, attn_probs)
            return output, attn_probs


class ColumnSelfAttention(nn.Module):
    """Compute self-attention over columns of a 2D input."""

    def __init__(
        self,
        embed_dim,
        num_heads,
        dropout=0.0,
        max_tokens_per_msa: int = 2 ** 16,
    ):
        super().__init__()

        self.num_heads = num_heads
        self.dropout = dropout
        self.head_dim = embed_dim // num_heads
        self.scaling = self.head_dim ** -0.5
        self.max_tokens_per_msa = max_tokens_per_msa

        self.k_proj = nn.Linear(embed_dim, embed_dim)
        self.v_proj = nn.Linear(embed_dim, embed_dim)
        self.q_proj = nn.Linear(embed_dim, embed_dim)

        self.out_proj = nn.Linear(embed_dim, embed_dim)
        self.dropout_module = nn.Dropout(dropout)

    def _batched_forward(
        self,
        x,
        self_attn_mask=None,
        self_attn_padding_mask=None,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        max_cols = max(1, self.max_tokens_per_msa // num_rows)
        outputs = []
        attns = []
        for start in range(0, num_cols, max_cols):
            output, attn = self(
                x[:, start : start + max_cols],
                self_attn_mask=self_attn_mask,
                self_attn_padding_mask=self_attn_padding_mask[:, :, start : start + max_cols]
                if self_attn_padding_mask is not None
                else None,
            )
            outputs.append(output)
            attns.append(attn)
        output = torch.cat(outputs, 1)
        attns = torch.cat(attns, 1)
        return output, attns

    def compute_attention_update(
        self,
        x,
        self_attn_mask=None,
        self_attn_padding_mask=None,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        if num_rows == 1:
            # if there is only 1 position, this is equivalent and doesn't break with padding
            attn_probs = torch.ones(
                self.num_heads,
                num_cols,
                batch_size,
                num_rows,
                num_rows,
                device=x.device,
                dtype=x.dtype,
            )
            output = self.out_proj(self.v_proj(x))
        else:
            q = self.q_proj(x).view(num_rows, num_cols, batch_size, self.num_heads, self.head_dim)
            k = self.k_proj(x).view(num_rows, num_cols, batch_size, self.num_heads, self.head_dim)
            v = self.v_proj(x).view(num_rows, num_cols, batch_size, self.num_heads, self.head_dim)
            q *= self.scaling

            attn_weights = torch.einsum("icnhd,jcnhd->hcnij", q, k)

            if self_attn_mask is not None:
                raise NotImplementedError
            if self_attn_padding_mask is not None:
                attn_weights = attn_weights.masked_fill(
                    self_attn_padding_mask.permute(2, 0, 1).unsqueeze(0).unsqueeze(3),
                    -10000,
                )

            attn_probs = attn_weights.softmax(-1)
            attn_probs = self.dropout_module(attn_probs)
            context = torch.einsum("hcnij,jcnhd->icnhd", attn_probs, v)
            context = context.contiguous().view(num_rows, num_cols, batch_size, embed_dim)
            output = self.out_proj(context)
        return output, attn_probs

    def forward(
        self,
        x,
        self_attn_mask=None,
        self_attn_padding_mask=None,
    ):
        num_rows, num_cols, batch_size, embed_dim = x.size()
        # if False and num_rows * num_cols > 2 ** 14 and not torch.is_grad_enabled():
        if (num_rows * num_cols) > self.max_tokens_per_msa and not torch.is_grad_enabled():
            return self._batched_forward(
                x,
                self_attn_mask,
                self_attn_padding_mask,
            )
        else:
            return self.compute_attention_update(x, self_attn_mask, self_attn_padding_mask)



import faiss
import re
import pandas as pd
import torch
import esm
import numpy as np
from Bio import SeqIO
from math import ceil
from ..model import MSARetrieveModel
from ..vector_construction import fasta2vec
from ..utils import TimeCounter


# 构建Faiss索引
def contruct_faiss_index(vectors, path, dim, measure, param):
	index = faiss.index_factory(dim, param, measure)
		
	if not index.is_trained:
		index.train(vectors)
	index.add(vectors)
	
	faiss.write_index(index, path)


if __name__ == '__main__':
	
	# model_name = "esm1b_t33_650M_UR50S"
	model_name = "RetrieveModel_scope2.08_d1_m1_t62_I663_B256_phase3"
	params = {"device_ids": [0, 1, 2, 3],
	          "model_path": f"../PretrainedModels/{model_name}.pt",
	          "fasta_path": f"/sujin/dataset/uniclust30/UniRef30_2020_06_100_200",
	          "save_path": f"/sujin/dataset/uniclust30/UniRef30_2020_06_100_200",
	          "batch_size": 64,
	          "log_name": None,
	          "cover": True}
	# fasta2vec(**params)
	# file_names = ['400_500']
	# for name in file_names:
	# 	params['fasta_path'] = f"/sujin/dataset/uniclust30/UniRef30_2020_06_{name}"
	# 	params['save_path'] = f"/sujin/dataset/uniclust30/UniRef30_2020_06_{name}"
	# 	fasta2vec(**params)

	
	name = '600_700'
	path = "/sujin/dataset/uniclust30/UniRef30_2020_06"

	vectors = np.load(f"{path}_{name}.npy")
	print(vectors.shape[0])

	n = ceil(vectors.shape[0] / 256)
	dim, measure = 1280, faiss.METRIC_L2
	param = f'IVF{n}, Flat'
	save_path = f"{path}_{name}_1.index"
	with TimeCounter("Constructing index..."):
		contruct_faiss_index(vectors, save_path, dim, measure, param)



from typing import List
import logging
import torch
from collections import OrderedDict
import numpy as np

logger = logging.getLogger(__name__)
ESM_1B = OrderedDict([
    ("<cls>", 0),
    ("<pad>", 1),
    ("<sep>", 2),
    ("<unk>", 3),
    ("L", 4),
    ("A", 5),
    ("G", 6),
    ("V", 7),
    ("S", 8),
    ("E", 9),
    ("R", 10),
    ("T", 11),
    ("I", 12),
    ("D", 13),
    ("P", 14),
    ("K", 15),
    ("Q", 16),
    ("N", 17),
    ("F", 18),
    ("Y", 19),
    ("M", 20),
    ("H", 21),
    ("W", 22),
    ("C", 23),
    ("X", 24),
    ("B", 25),
    ("U", 26),
    ("Z", 27),
    ("O", 28),
    (".", 29),
    ("-", 30),
    ("<null_1>", 31),
    ("<mask>", 32)])


class Tokenizer():
    def __init__(self):
        self.vocab = ESM_1B
        self.tokens = list(self.vocab.keys())
        assert self.start_token in self.vocab and self.stop_token in self.vocab

    @property
    def vocab_size(self) -> int:
        return len(self.vocab)

    @property
    def start_token(self) -> str:
        return "<cls>"

    @property
    def stop_token(self) -> str:
        return "<sep>"

    @property
    def mask_token(self) -> str:
        if "<mask>" in self.vocab:
            return "<mask>"
        else:
            raise RuntimeError(f"{self._vocab_type} vocab does not support masking")

    def tokenize(self, text: str) -> List[str]:
        return [x for x in text]

    def convert_token_to_id(self, token: str) -> int:
        """ Converts a token (str/unicode) in an id using the vocab. """
        try:
            return self.vocab[token]
        except KeyError:
            raise KeyError(f"Unrecognized token: '{token}'")

    def convert_tokens_to_ids(self, tokens: List[str]) -> List[int]:
        return [self.convert_token_to_id(token) for token in tokens]

    def convert_id_to_token(self, index: int) -> str:
        """Converts an index (integer) in a token (string/unicode) using the vocab."""
        try:
            return self.tokens[index]
        except IndexError:
            raise IndexError(f"Unrecognized index: '{index}'")

    def convert_ids_to_tokens(self, indices: List[int]) -> List[str]:
        return [self.convert_id_to_token(id_) for id_ in indices]

    def convert_tokens_to_string(self, tokens: str) -> str:
        """ Converts a sequence of tokens (string) in a single string. """
        return ''.join(tokens)

    def add_special_tokens(self, token_ids: List[str]) -> List[str]:
        """
        Adds special tokens to the a sequence for sequence classification tasks.
        A BERT sequence has the following format: [CLS] X [SEP]
        """
        cls_token = [self.start_token]
        sep_token = [self.stop_token]
        return cls_token + token_ids + sep_token

    def single_encode(self, text: str) -> np.ndarray:
        tokens = self.tokenize(text)
        tokens = self.add_special_tokens(tokens)
        token_ids = self.convert_tokens_to_ids(tokens)
        return np.array(token_ids, np.int64)

    def batch_encode(self, seqs, padding=False):
        batch = [self.single_encode(seq) for seq in seqs]
        pad_id = self.convert_token_to_id('<pad>')
        lengths = [len(seq) for seq in seqs]
        input_ids = torch.from_numpy(self.pad_sequences(batch, pad_id)) if padding else batch

        out = {'input_ids': input_ids, 'lengths': lengths}
        return out

    # 进行序列的padding操作
    def pad_sequences(self, sequences, constant_value=0, dtype=None) -> np.ndarray:
        batch_size = len(sequences)
        shape = [batch_size] + np.max([seq.shape for seq in sequences], 0).tolist()

        if dtype is None:
            dtype = sequences[0].dtype

        if isinstance(sequences[0], np.ndarray):
            array = np.full(shape, constant_value, dtype=dtype)
        elif isinstance(sequences[0], torch.Tensor):
            array = torch.full(shape, constant_value, dtype=dtype)

        for arr, seq in zip(array, sequences):
            arrslice = tuple(slice(dim) for dim in seq.shape)
            arr[arrslice] = seq

        return array



import torch
from . import esm
from .PretrainedModels.config import load_model_architecture
from torch.nn.functional import gelu


class MSARetrieveModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.esm1b = load_model_architecture(name="esm-1b")

        # 去掉不用的参数的梯度
        for param in self.esm1b.lm_head.parameters():
            param.requires_grad = False
        for param in self.esm1b.contact_head.parameters():
            param.requires_grad = False
            
        self.layer = self.esm1b.num_layers
        self.dim = self.esm1b.args.embed_dim

    def forward(self, inputs, lengths=None, use_checkpoint=False):
        outputs = self.esm1b(inputs, repr_layers=[self.layer], use_checkpoint=use_checkpoint)

        if lengths:
            outputs = outputs["representations"][self.layer]
            vec_out = outputs[0, 1: lengths[0]+1].mean(dim=0, keepdim=True)
            for i, l in enumerate(lengths[1:]):
                vec = outputs[i+1, 1: l+1].mean(dim=0, keepdim=True)
                vec_out = torch.cat((vec_out, vec), dim=0)

        else:
            vec_out = outputs["representations"][self.layer][:, 0, :]

        out_dict = {'repr': vec_out}

        return out_dict


class ThresholdModel(torch.nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.fc1 = torch.nn.Linear(dim, dim * 2)
        self.fc2 = torch.nn.Linear(dim * 2, dim * 2)
        self.out = torch.nn.Linear(dim * 2, 1)

    def forward(self, inputs):
        x1 = gelu(self.fc1(inputs))
        x2 = gelu(self.fc2(x1))
        t = gelu(self.out(x2)).squeeze(-1)
        return t


if __name__ == '__main__':
    model = MSARetrieveModel()
    print(model)
